#!/usr/bin/env python
#pylint: disable=C0302

"""
@copyright: Copyright (C) 2008, 2016 Oracle and/or its affiliates. All rights reserved.

@license:   See the file COPYING for redistribution information.

@summary:   This module contains all the Oracle Storage Connect API
            interfaces.
"""

__docformat__ = 'epytext en'

import sys
import os
import socket
import gettext
import subprocess
import platform

import OSCCache
from PyEnum import PyEnum

#
# Import the module that provides the appropriate OS specific functions
# based on the operating system we are running on.
#

SAN = { 'Linux': 'san_linux',
        'SunOS': 'san_solaris' }

san = __import__(SAN[platform.system()], globals(), locals(), [])

__all__ = ["IPlugin",

           "IStorageArrayPlugin",
           "IFileSystemPlugin",

           "IPluginException",

           "NoSuchOperationEx",
           "OperationFailedEx",
           "MissingKeyEx",
           "ValueFormatEx",
           "StorageArrayLicenseEx",
           "TargetDiscoveryEx",
           "LoginFailedEx",
           "LogoutFailedEx",
           "RefreshFailedEx",
           "ListFailedEx",
           "CreateSnapFailedEx",
           "ListSnapFailedEx",
           "SnapRestoreNotSafeEx",
           "CloneFailedEx",
           "InvalidStorageArrayEx",
           "InvalidValueEx",
           "StorageElementBusyEx",
           "PermissionDeniedEx",
           "OperationPreReqNotMetEx",
           "StorageNameRequiredEx",
           "InvalidFSTypeEx",
           "FileSystemBusyEx",
           "CommandMissingEx",
           "FileSystemAlreadyMountedEx",

           "ABILITY_TYPES",
           "SE_STATES",
           "SE_STATES_TEXT",
           "FS_STATES",
           "FS_STATES_TEXT",
           "BACKING_DEVICE_TYPES",
           "BACKING_DEVICE_TYPES_TEXT",
          ]

_ = gettext.gettext



ABILITY_TYPES = PyEnum("UNSUPPORTED             = 0x00",
                       "OFFLINE                 = 0x01",
                       "ONLINE                  = 0x02",
                       "NO                      = 0x04",
                       "YES                     = 0x08",
                       "INVALID                 = 0xFF",
                      )
"""
Enum representing the possible values a plugin/storage server
capability can be set to.
"""

SE_STATES = PyEnum("UNKNOWN                     = 0x00",
                   "OFFLINE                     = 0x01",
                   "BUSY_CREATE                 = 0x02",
                   "BUSY_CLONE                  = 0x03",
                   "BUSY_COPY                   = 0x04",
                   "BUSY_CHILDREN               = 0x05",
                   "ERROR                       = 0x06",
                   "ONLINE                      = 0x07",
                  )
"""
Enum representing all the possible Storage Element states.
"""

SE_STATES_TEXT = {SE_STATES.UNKNOWN:            "Unknown",
                  SE_STATES.OFFLINE:            "Offline",
                  SE_STATES.BUSY_CREATE:        "Busy (Creating)",
                  SE_STATES.BUSY_CLONE:         "Busy (Cloning)",
                  SE_STATES.BUSY_COPY:          "Busy (Copying)",
                  SE_STATES.BUSY_CHILDREN:      "Busy (Have children)",
                  SE_STATES.ERROR:              "Error",
                  SE_STATES.ONLINE:             "Online",
                 }
"""
Dictionary used to lookup the text description associated with the
state of a Storage Element.
"""

FS_STATES = PyEnum("UNKNOWN                     = 0x00",
                   "UNMOUNTED                   = 0x01",
                   "MOUNTED                     = 0x02",
                   "MOUNTED_READONLY            = 0x03",
                   "ERROR                       = 0x04",
                  )
"""
Enum representing all the possible File System states.
"""

FS_STATES_TEXT = {FS_STATES.UNKNOWN:            "Unknown",
                  FS_STATES.UNMOUNTED:          "Not mounted",
                  FS_STATES.MOUNTED:            "Mounted",
                  FS_STATES.MOUNTED_READONLY:   "Mounted Read Only",
                  FS_STATES.ERROR:              "Error",
                 }
"""
Dictionary used to lookup the text description associated with the
state of a File System.
"""

BACKING_DEVICE_TYPES = PyEnum("UNSUPPORTED                 = 0x00",
                              "DEVICE_SINGLE               = 0x01",
                              "DEVICE_MULTI                = 0x02",
                              "PLUGIN_SINGLE               = 0x03",
                              "PLUGIN_MULTI                = 0x04",
                              "INVALID                     = 0x05",
                             )
"""
Enum representing all the possible File System states.
"""

BACKING_DEVICE_TYPES_TEXT = {BACKING_DEVICE_TYPES.UNSUPPORTED:     "File system creation is not supported",
                             BACKING_DEVICE_TYPES.DEVICE_SINGLE:   "Single block device",
                             BACKING_DEVICE_TYPES.DEVICE_MULTI:    "Multiple block devices",
                             BACKING_DEVICE_TYPES.PLUGIN_SINGLE:   "Single plugin defined value",
                             BACKING_DEVICE_TYPES.PLUGIN_MULTI:    "Multiple plugin defined values",
                             BACKING_DEVICE_TYPES.INVALID:         "Invalid",
                            }
"""
Dictionary used to lookup the text description associated with the
state of a File System.
"""

class IPlugin(object):
    """
    The base interface for various Storage Connect plug-in interfaces.

    The class tree of the Storage Connect API is as follows:
    G{classtree: IPlugin}

    @note: All logging should be done through the L{logger <IPlugin.logger>}
           instance set in the L{IPlugin<IPlugin>} class.
    """

    name = "IPlugin"
    """
    @cvar: Class name.
    @type: C{str}
    """

    plugin_api_version = ["1", "2", "7"]
    """
    @cvar: The over arching API version for the plugin interface.
    @type: C{list} of C{str}
    """

    required_api_vers = None
    """
    @cvar: The minimum API version required by the plugin.
    @type: C{list} of C{str}
    """

    extended_api_version = None
    """
    @cvar: The minimum extended API version implemented by the plugin.
    @type: C{list} of C{str}
    """

    logger = None
    """
    @cvar: Logging instance to be used by the plugin.
    @type: C{logging} object
    """

    iscsiadm = "/sbin/iscsiadm"
    """
    @cvar: Location of the iscsiadm executable.
    @type: C{str}
    """

    plugin_type = "inv"
    """
    @cvar: Specific type of plugin.
    @type: C{str}
    """

    plugin_name = ""
    """
    @cvar: The name of the plugin - Name used to populate the user
           interface drop down list.
    @type: C{str}
    """

    vendor_name = ""
    """
    @cvar: Name of the Vendor supplying the plugin.
    @type: C{str}
    """

    plugin_version = ""
    """
    @cvar: Version of the plugin.
    @type: C{str}
    """

    ss_extra_info_help = "None"
    """
    @cvar: The help text that will be displayed for the extra info
           to the user when creating the Storage Server.
    @type: C{str}
    """

    se_extra_info_help = "None"
    """
    @cvar: The help text that will be displayed for the extra info
           to the user when creating a Storage Element.
    @type: C{str}
    """

    fs_extra_info_help = "None"
    """
    @cvar: The help text that will be displayed for the extra info
           to the user when creating a File System.
    @type: C{str}
    """

    file_extra_info_help = "None"
    """
    @cvar: The help text that will be displayed for the extra info
           to the user when creating a File.
    @type: C{str}
    """

    plugin_ability = {"snapshot":               ABILITY_TYPES.INVALID,
                      "custom_snap_name":       ABILITY_TYPES.INVALID,
                      "snap_is_sync":           ABILITY_TYPES.INVALID,
                      "clone":                  ABILITY_TYPES.INVALID,
                      "custom_clone_name":      ABILITY_TYPES.INVALID,
                      "clone_is_sync":          ABILITY_TYPES.INVALID,
                      "resize":                 ABILITY_TYPES.INVALID,
                      "resize_is_sync":         ABILITY_TYPES.INVALID,
                      "splitclone":             ABILITY_TYPES.INVALID,
                      "splitclone_is_sync":     ABILITY_TYPES.INVALID,
                      "splitclone_while_open":  ABILITY_TYPES.INVALID,
                      "snapclone":              ABILITY_TYPES.INVALID,
                      "snapclone_is_sync":      ABILITY_TYPES.INVALID,
                      "require_storage_name":   ABILITY_TYPES.INVALID,
                      "backing_device_type":    BACKING_DEVICE_TYPES.INVALID,
                      "access_control":         ABILITY_TYPES.INVALID,
                      "max_access_entries":     0,
                     }
    """
    @cvar: Capabilities that this specific plugin support.
           L{Version<plugin_api_version>} of the API have the following
           fields::

                Name                                   Comment
            --------------------------------------------------------------------
            "snapshot":                   # Set to ONLINE/OFFLINE if the plugin
                                          #   can create snapshots, NO if not.
            "custom_snap_name":           # Set this to YES if a custom name can
                                          #   be given to the snapshot.
            "snap_is_sync":               # Set this to YES if creating the
                                          #   snapshot is synchronously.
            "clone":                      # Set to ONLINE/OFFLINE if the plugin
                                          #   can create clones, NO if not.
            "custom_clone_name":          # Set this to YES if a custom name can
                                          #   be given to the new clone.
            "clone_is_sync":              # Set this to YES if creating the
                                          #   clone is synchronously.
            "resize":                     # Set to ONLINE/OFFLINE if the plugin
                                          #   can resize Storage Elements, NO
                                          #   if not.
            "resize_is_sync":             # Set this to YES if resizing the
                                          #   Storage Element is synchronously.
            "splitclone":                 # Set to ONLINE/OFFLINE if the plugin
                                          #   can split clones, NO if not.
            "splitclone_is_sync":         # Set this to YES if splitting the
                                          #   clones is synchronously.
            "splitclone_while_open":      # Set this to YES if the plugin allows
                                          #   the clones to open when splitting.
            "snapclone":                  # Set to ONLINE/OFFLINE if the plugin
                                          #   can create a clone from a snapshot,
                                          #   NO if not.
            "snapclone_is_sync":          # Set this to YES if cloning from a
                                          #   snapshot is synchronously.
            "require_storage_name":       # Set this to YES if the plugin require
                                          #   the user to supply a Storage Name
                                          #   (normally used by storage
                                          #    concentrators and appliances)
            "backing_device_type":        # This is for File System plugins ONLY,
                                          #   it defines what type of backing device
                                          #   the file system use when creating a
                                          #   file system.
            "access_control":             # Set this to YES if the plugin support
                                          #   access control groups.
            "max_access_entries":         # Set this to the maximum number any one
                                          #   access group can contain.
    @type: C{dict}
    """

    dev_path_prefix = "/dev/mapper"
    """
    @cvar: This is the device path prefix that plug-ins should use when filling
           in the path in the L{SERecord<__SERecord_fields__>}.
    @type: C{str}
    """

    dev_path_by_id = "/dev/disk/by-id"
    """
    @cvar: This is the device path prefix by-id that plug-ins should use when
           filling in the path in the L{SERecord<__SERecord_fields__>}.
    @type: C{str}
    """

    dev_name_regex = "^[0-9a-zA-Z]+"
    """
    @cvar: This is the device name regex that plug-ins should use when scanning
           through the devices to fill in the path in
           L{SERecord<__SERecord_fields__>}.
    @type: C{str}
    """

    SANStorage = "SAN"
    """
    @cvar: Used to designate that the plugin can handle all types of SCSI
           storage (FC, FCoE, SAS, SATA, etc.) except iSCSI.
    @type: C{str}
    """

    iSCSIStorage = "iSCSI"
    """
    @cvar: Used to designate that the plugin can handle iSCSI storage.
    @type: C{str}
    """

    FileSysStorage = "FileSys"
    """
    @cvar: Used to designate that the plugin handle File System storage.
    @type: C{str}
    """

    NetworkFileSystem = "NetworkFS"
    """
    @cvar: Used to designate that the plugin is for a network file system.
    @type: C{str}
    """

    LocalFileSystem = "LocalFS"
    """
    @cvar: Used to designate that the plugin is for a local file system.
    @type: C{str}
    """

    __access_descriptor__ = {"access_host":        "",
                             "access_port":        0,
                             "username":           "",
                             "passwd":             "",
                             }
    """
    @ivar: Access descriptors. All the known fields and if the key is mandatory
           are listed in L{Access descriptor fields<__access_descriptor_fields__>}.
    @type: C{dict}
    """

    __access_descriptor_fields__ = {"access_host":        (str,       False),
                                    "access_port":        (int,       False),
                                    "username":           (str,       False),
                                    "passwd":             (str,       False),
                                   }
    
    """
    @cvar: Field list for the access_descriptor. 
    @type: C{dict}
    """

    __access_grp__ = {"grp_name":               "",
                      "grp_entries":            [""],
                      "grp_modes":              [""],
                     }
    """
    @ivar: Access group. All the known fields and if the key is mandatory
           are listed in L{Access Group fields<__access_grp_fields__>}.
    @type: C{dict}
    """

    __access_grp_fields__ = {"grp_name":        (str,       True),
                             "grp_entries":     (list,      True),
                             "grp_modes":       (list,      False),
                            }
    """
    @cvar: Field list for the access group. L{Version<plugin_api_version>} of
           the API have the following fields::

                Name               Type                    Comment
            --------------------------------------------------------------------
            "grp_name":         str         # Name of the access group.
            "grp_entries":      list        # This is the access control entries
                                            #   for the group. For example,
                                            #   SAN would have a list of WWNs,
                                            #   iSCSI, a list of initiator names
                                            #   and for network file systems, a
                                            #   list of host names.
            "grp_modes":        list        # Optional list of modes associated
                                            #   with the access group, for
                                            #   example the export mode for a
                                            #   network file system.
    @type: C{dict}
    """

    __vol_group__ = {"vol_name":                "",
                     "vol_total_sz":            0,
                     "vol_used_sz":             0,
                     "vol_free_sz":             0,
                     "vol_alloc_sz":            0,
                     "vol_desc":                "",
                    }
    """
    @ivar: Volume group. All the known fields and if the key is mandatory
           are listed in L{Volume group fields<__vol_group_fields__>}.
    @type: C{dict}
    """

    __vol_group_fields__ = {"vol_name":         (str,       True),
                            "vol_total_sz":     (int,       True),
                            "vol_used_sz":      (int,       True),
                            "vol_free_sz":      (int,       True),
                            "vol_alloc_sz":     (int,       False),
                            "vol_desc":         (str,       False),
                           }
    """
    @cvar: Field list for the volume group. L{Version<plugin_api_version>} of
           the API have the following fields::

                Name               Type                    Comment
            --------------------------------------------------------------------
            "vol_name":         str         # Name of the volume group.
            "vol_total_sz":     int         # Total size of the volume group.
            "vol_used_sz":      int         # Currently used space.
            "vol_free_sz":      int         # Currently available space.
            "vol_alloc_sz":     int         # Currently allocated space.
            "vol_desc":         str         # Volume group description.
    @type: C{dict}
    """

    __qos_vals__ = {"priority":                 0,
                    "value":                    "",
                   }
    """
    @ivar: Quality-of-Service. All the known fields and if the key is mandatory
           are listed in L{QoS fields<__qos_vals_fields__>}.
    @type: C{dict}
    """

    __qos_vals_fields__ = {"priority":          (int,       True),
                           "value":             (str,       True),
                          }
    """
    @cvar: Field list for the qos_vals. L{Version<plugin_api_version>} of the
           API have the following fields::

                Name               Type                    Comment
            --------------------------------------------------------------------
            "priority":         int         # Priority of the QoS value,
                                            #   zero (0) being the highest
                                            #   priority. It is not required for
                                            #   the values to follow numerically
            "value":            str         # Quality-of-Service name as known
                                            #   by the Storage Server.
    @type: C{dict}
    """

    __SSRecord__ = {"name":                     "",
                    "uuid":                     "",
                    "storage_server_id":        "",
                    "storage_type":             "",
                    "access_host":              "",
                    "access_port":              "",
                    "username":                 "",
                    "passwd":                   "",
                    "chap":                     False,
                    "admin_host":               "",
                    "admin_user":               "",
                    "admin_passwd":             "",
                    "netdevs":                  [""],
                    "storage_name":             "",
                    "status":                   "",
                    "total_sz":                 0,
                    "used_sz":                  0,
                    "free_sz":                  0,
                    "alloc_sz":                 0,
                    "access_grps":              [__access_grp__],
                    "vol_groups":               [__vol_group__],
                    "storage_id":               [""],
                    "storage_desc":             "",
                    "extra_info":               "",
                    "access_descriptors":       [__access_descriptor__],
                   }
    """
    @cvar: Storage Server record. All the known fields and if the key is
           mandatory are listed in L{SSRecord<__SSRecord_fields__>}.
    @type: C{dict}
    """

    __SSRecord_fields__ = {"name":              (str,       True),
                           "uuid":              (str,       True),
                           "storage_server_id": (str,       False),
                           "storage_type":      (str,       True),
                           "access_host":       (str,       False),
                           "access_port":       (int,       False),
                           "username":          (str,       False),
                           "passwd":            (str,       False),
                           "chap":              (bool,      False),
                           "admin_host":        (str,       True),
                           "admin_user":        (str,       True),
                           "admin_passwd":      (str,       True),
                           "netdevs":           (list,      False),
                           "storage_name":      (str,       False),
                           "status":            (str,       True),
                           "total_sz":          (int,       True),
                           "used_sz":           (int,       True),
                           "free_sz":           (int,       True),
                           "alloc_sz":          (int,       False),
                           "access_grps":       (list,      True),
                           "vol_groups":        (list,      False),
                           "storage_id":        (list,      False),
                           "storage_desc":      (str,       True),
                           "extra_info":        (str,       False),
                           "access_descriptors": (list,      False),
                          }
    """
    @cvar: Field list for the Storage Server record.
           L{Version<plugin_api_version>} of the API have the following fields::

                Name               Type                    Comment
            --------------------------------------------------------------------
            "name":                str      # Name of the Storage Server as
                                            #   shown in the Oracle VM Manager.
            "uuid":                str      # UUID string of the Storage Server
                                            #   Record.
            "storage_server_id":   str      # Storage Server unique identifier.
                                            #   This is for use solely by the
                                            #   plug-in to uniquely identify
                                            #   the Storage Server.
            "storage_type":        str      # Storage type: SAN, iSCSI or
                                            #   FileSys.
            "access_host":         str      # Host/IP address used to access the
                                            #   storage.
            "access_port":         int      # Port number to access the storage.
            "username":            str      # Storage server username.
            "passwd":              str      # Storage server password.
            "chap":                bool     # Does storage server require CHAP?
            "admin_host":          str      # Host/IP address to admin the
                                            #   storage.
            "admin_user":          str      # Storage administrator user name.
            "admin_passwd":        str      # Storage administrator password.
            "netdevs":             list     # Network device(s) used to access
                                            #   the storage.
            "storage_name":        str      # Actual name of the Storage Server.
            "status":              str      # Status of the Storage Server.
            "total_sz":            int      # Total size in bytes.
            "used_sz":             int      # Actual space already used.
            "free_sz":             int      # Available space in bytes.
            "alloc_sz":            int      # Allocated/reserved space in bytes.
            "access_grps":         list     # List of access groups. See
                                            #   __access_grp_fields__
            "vol_groups":          list     # List of vol_groups. See
                                            #   __vol_group_fields__
            "storage_id":          str      # Storage identifier (iSCSI target
                                            #   name).
            "storage_desc":        str      # Storage description.
            "extra_info":          str      # Extra info (open format).
            "access_descriptor":   list     # List of access descriptors. See
                                            #   __access_descriptor_fields__

            NOTE: Depending on the storage types some fields would not be in the
                  dict for instance SAN would not have the username, passwd etc.
                  also if the start method does not return a updated record with
                  the storage_id set it would not be passed in subsequent
                  operations.
    @type: C{dict}
    """
    __SERecord__ = {"se_type":                  "",
                    "ss_uuid":                  __SSRecord__["uuid"],
                    "name":                     "",
                    "uuid":                     "",
                    "page83_id":                "",
                    "id":                       "",
                    "vendor":                   "",
                    "product_id":               "",
                    "path":                     [""],
                    "array_path":               "",
                    "size":                     0,
                    "status":                   "",
                    "vol_group_name":           "",
                    "access_grp_names":         [""],
                    "qos":                      "",
                    "state":                    SE_STATES.UNKNOWN,
                    "extra_info":               "",
                   }
    """
    @cvar: Storage Element record. All the known fields and if the key is
           mandatory are listed in L{SERecord<__SERecord_fields__>}.
    @type: C{dict}
    """

    __SERecord_fields__ = {"se_type":           (str,       True),
                           "ss_uuid":           (str,       True),
                           "name":              (str,       True),
                           "uuid":              (str,       True),
                           "page83_id":         (str,       True),
                           "id":                (str,       False),
                           "vendor":            (str,       False),
                           "product_id":        (str,       False),
                           "path":              (list,      True),
                           "array_path":        (str,       False),
                           "size":              (int,       True),
                           "status":            (str,       True),
                           "vol_group_name":    (str,       False),
                           "access_grp_names":  (list,      True),
                           "qos":               (str,       False),
                           "state":             (SE_STATES, True),
                           "extra_info":        (str,       False),
                          }
    """
    @cvar: Field list for the Storage Element record.
           L{Version<plugin_api_version>} of the API have the following fields::

                Name               Type                    Comment
            --------------------------------------------------------------------
            "se_type":          str         # See __SETypes__.
            "ss_uuid":          str         # UUID for the owning SSRecord.
            "name":             str         # User friendly name.
            "uuid":             str         # UUID string for the device.
            "page83_id":        str         # SCSI Page 83 unique LUN ID.
            "id":               str         # ID for the SE (LUN/Snap ID).
            "vendor":           str         # Vendor as seen by a SCSI INQ.
            "product_id":       str         # Product as seen by a SCSI INQ.
            "path":             list        # System path(s) to the device.
            "array_path":       str         # Path on the array for the device.
            "size":             int         # Size in bytes.
            "status":           str         # Returned status of the SE.
            "vol_group_name":   str         # Volume group this SE belongs to.
            "access_grp_names": list        # List of access group names.
            "qos":              str         # Quality-of-Service for the SE.
            "state":            SE_STATES   # Current state of the SE.
            "extra_info":       str         # Extra info (open format).

            NOTE: Depending on the storage types some fields would not be in the
                  dict for instance SAN would not have the username, passwd etc.
    @type: C{dict}
    """

    LUNType = "LUN"
    """
    @cvar: Used to designate that the SE record represent a LUN.
    @type: C{str}
    """

    GhostLUNType = "GhostLUN"
    """
    @cvar: Used to designate that the SE record represent a Ghost LUN.
    @type: C{str}
    """

    SnapType = "Snapshot"
    """
    @cvar: Used to designate that the SE record represent a Snapshot.
    @type: C{str}
    """

    SnapCloneType = "SnapClone"
    """
    @cvar: Used to designate that the SE record represent a clone from a
           snapshot.
    @type: C{str}
    """

    __SETypes__ = [LUNType,
                   GhostLUNType,
                   SnapType,
                   SnapCloneType,
                   ]
    """
    @cvar: Known Storage Element types. A
           L{SERecord<__SERecord_fields__>} must be of these types.
           L{Version <plugin_api_version>} of the API define the following
           types::

                Type                              Comment
            --------------------------------------------------------------------
            LUNType         # Normal block device LUN.
            GhostLUNType    # Ghost LUN.
            SnapType        # Snapshot created for a LUN, File system or File.
            SnapCloneType   # Clone created from a snapshot. This is a special
                            #   case since it can become a full LUN in the
                            #   future. We need to keep track as on most arrays
                            #   the snap is locked until the snap clone is
                            #   deleted or split from the snapshot.
    @type: C{list}
    """

    __FSRecord__ = {"access_path":              "",
                    "ss_uuid":                  __SSRecord__["uuid"],
                    "name":                     "",
                    "mount_options":            [""],
                    "array_path":               "",
                    "uuid":                     "",
                    "size":                     0,
                    "status":                   "",
                    "state":                    FS_STATES.UNKNOWN,
                    "access_grp_names":         [""],
                    "qos":                      "",
                    "backing_device":           "",
                    "extra_info":               "",
                   }
    """
    @cvar: File System record. All the known fields and if the key is
           mandatory are listed in L{FSRecord<__FSRecord_fields__>}.
    @type: C{dict}
    """

    __FSRecord_fields__ = {"access_path":       (str,       False),
                           "ss_uuid":           (str,       True),
                           "name":              (str,       True),
                           "mount_options":     (list,      False),
                           "array_path":        (str,       False),
                           "uuid":              (str,       True),
                           "size":              (int,       True),
                           "status":            (str,       True),
                           "state":             (FS_STATES, True),
                           "access_grp_names":  (list,      True),
                           "qos":               (str,       False),
                           "backing_device":    (str,       False),
                           "extra_info":        (str,       False),
                          }
    """
    @cvar: Field list for the File System record. L{Version<plugin_api_version>}
           of the API have the following fields::

                Name               Type                    Comment
            --------------------------------------------------------------------
            "access_path":      str         # Access path for the file system,
                                            #   this is only used for networked
                                            #   file systems.
            "ss_uuid":          str         # UUID of the SS Record owning this
                                            #   file system.
            "name":             str         # User friendly name of the file
                                            #   system.
            "mount_options":    list        # List of default mount options for
                                            #   the file system.
            "array_path":       str         # Path on storage server for the
                                            #   file system.
            "uuid":             str         # UUID of the the file system.
            "size":             int         # Size, returned, supplied for
                                            #   resizing.
            "status":           str         # Status of the file system for
                                            #   display.
            "state":            FS_STATES   # State of the file system.
            "access_grp_names": list        # List of access group names the
                                            #   file system is (or will be)
                                            #   exported to.
            "qos":              str         # Quality-of-Service for the FS.
            "backing_device":   str/list    # This is used to specify what type
                                            #   of device the File System
                                            #   requires.
            "extra_info":       str         # Extra info (open format extra
                                            #   info).

            NOTE: Depending on the file system type some fields would not be in
                  the dict.
    @type: C{dict}
    """

    __BackingDeviceRecord__ = {"name":              "",
                               "total_sz":          0,
                               "free_sz":           0,
                               "value":             "",
                               "description":       "",
                              }
    """
    @cvar: Backing device record. All the known fields and if the key is
           mandatory are listed in L{BackingDeviceRecord<__BackingDeviceRecord_fields__>}.
    @type: C{dict}
    """

    __BackingDeviceRecord_fields__ = {"name":           (str,       True),
                                      "total_sz":       (int,       True),
                                      "free_sz":        (int,       True),
                                      "value":          (list,      True),
                                      "description":    (str,       False),
                                     }
    """
    @cvar: Field list for the backing device records.
           L{Version<plugin_api_version>} of the API have the following
           fields::

                Name               Type                    Comment
            --------------------------------------------------------------------
            "name":             str         # Name of the backing device.
            "total_sz":         int         # Total size of the backing device.
            "free_sz":          int         # Available free size in the backing
                                            #   device.
            "value":            str         # Value that need to be passed back
                                            #   to the plugin to use this backing
                                            #   device.
            "description":      str         # User friendly description for the
                                            #   backing device that will be.
    @type: C{dict}
    """

    __MountRecord__ = {"uuid":                  "",
                       "fs_uuid":               __FSRecord__["uuid"],
                       "mount_point":           "",
                       "options":               [""],
                       "shared_path":           "",
                       "status":                "",
                       "extra_info":            "",
                      }
    """
    @cvar: Mount point record. All the known fields and if the key is
           mandatory are listed in L{MountRecord<__MountRecord_fields__>}.
    @type: C{dict}
    """

    __MountRecord_fields__ = {"uuid":           (str,       True),
                              "fs_uuid":        (str,       True),
                              "mount_point":    (str,       True),
                              "options":        (list,      False),
                              "status":         (str,       True),
                              "extra_info":     (str,       False),
                             }
    """
    @cvar: Field list for the Mount point records.
           L{Version<plugin_api_version>} of the API have the following
           fields::

                Name               Type                    Comment
            --------------------------------------------------------------------
            "uuid":             str         # UUID of this mount point.
            "fs_uuid":          str         # UUID of the File System owning
                                            #   this mount point.
            "mount_point":      str         # Path file system should be or is
                                            #   mounted on.
            "options":          list        # List of mount options for this
                                            #   specific mount point.
            "status":           str         # Status of the mount point for
                                            #   display.
            "extra_info":       str         # Extra info (open format extra
                                            #   info).
    @type: C{dict}
    """

    __FileRecord__ = {"fr_type":                "",
                      "file_path":              "",
                      "ondisk_sz":              0,
                      "file_sz":                0,
                      "fs_uuid":                __FSRecord__["uuid"],
                      "name_pattern":           "",
                      "snap_name":              "",
                      "extra_info":             "",
                     }
    """
    @cvar: File record. All the known fields and if the key is
           mandatory are listed in L{FileRecord<__FileRecord_fields__>}.
    @type: C{dict}
    """

    __FileRecord_fields__ = {"fr_type":         (str,       True),
                             "file_path":       (str,       True),
                             "ondisk_sz":       (int,       True),
                             "file_sz":         (int,       True),
                             "fs_uuid":         (str,       True),
                             "name_pattern":    (str,       False),
                             "snap_name":       (str,       False),
                             "extra_info":      (str,       False),
                            }
    """
    @cvar: Field list for the File System record. L{Version<plugin_api_version>}
           of the API have the following fields::

                Name               Type                    Comment
            --------------------------------------------------------------------
            "fr_type":          str         # See __FRTypes__.
            "file_path":        str         # Absolute path to the file.
            "fs_uuid":          str         # UUID of the file system containing
                                            #   the file.
            "name_pattern":     str         # This would optionally contain the
                                            #   pattern the file or directory
                                            #   name should match when listing.
            "snap_name":        str         # Name of the snap.
            "extra_info":       str         # Extra info (open format).

            NOTE: Depending on the file system type some fields would not be in
                  the dict.
    @type: C{dict}
    """

    FileType = "File"
    """
    @cvar: Used to designate that the FR record represent a file.
    @type: C{str}
    """

    DirType = "Directory"
    """
    @cvar: Used to designate that the FR record represent a directory.
    @type: C{str}
    """

    SymLinkType = "SymbolicLink"
    """
    @cvar: Used to designate that the FR record represent a symbolic link.
    @type: C{str}
    """

    CharDevType = "CharDevice"
    """
    @cvar: Used to designate that the FR record represent a character device.
    @type: C{str}
    """

    BlkDevType = "BlockDevice"
    """
    @cvar: Used to designate that the FR record represent a block device.
    @type: C{str}
    """

    FifoType = "FIFOFile"
    """
    @cvar: Used to designate that the FR record represent a FIFO file.
    @type: C{str}
    """

    SockFileType = "SocketFile"
    """
    @cvar: Used to designate that the FR record represent a socket file.
    @type: C{str}
    """

    __FRTypes__ = [FileType,
                   DirType,
                   SymLinkType,
                   CharDevType,
                   BlkDevType,
                   FifoType,
                   SockFileType,
                   SnapType,
                   SnapCloneType,
                  ]
    """
    @cvar: Known File record types. A
           L{FileRecord<__FileRecord_fields__>} must be of these types.
           L{Version<plugin_api_version>} of the API define the following
           types::

                Type                              Comment
            --------------------------------------------------------------------
            FileType        # Normal file.
            DirType         # Directory.
            SymLinkType     # Symbolic link.
            CharDevType     # Character device file.
            BlkDevType      # Block device file.
            FifoType        # FIFO pipe file.
            SocketType      # Socket file.
            SnapType        # Snapshot / RefLink created of the file.
            SnapCloneType   # Clone created from a snapshot. This is a special
                            #   case since in some file systems the Snapshot
                            #   must stay around as long as there is a clone
                            #   based on it. It is only used when the file
                            #   system have this restriction.
    @type: C{list}
    """

    __osc_install_dir__ = "/opt/storage-connect"
    """
    @cvar: The Oracle Storage Connect installation directory.
    @type: C{str}
    """

    __plugin_top_dir__ = "%s/plugins" % __osc_install_dir__
    """
    @cvar: Top directory under which all Storage Connect plugins will be
           installed.
    @type: C{str}
    """

    __plugin_types__ = ["isa", "ifs"]
    """
    @cvar: Known Oracle Storage Connect plugin types.
    @type: C{list}
    """

    __plugin_temp_dir__ = "/var/tmp/storage-connect"
    """
    @cvar: Directory where temporrary files woud be created under.
    @type: C{str}
    """

    __generic_plugin__ = False
    """
    @cvar: Flag if this is a generic Plugin - Vendor implemented plugins
           should NOT change this!
    @type: C{bool}
    """

    __cluster_required__ = False
    """
    @cvar: Flag if this Plugin requires the cluster to be up - Vendor
           implemented plugins should NOT change this!
    @type: C{bool}
    """

    cache = OSCCache.OSCDefaultCache()
    """
    @cvar: Caching that can be used by the plugin.
    @type: C{cache} object
    """

    def __new__(cls, *args, **kwargs):
        """
        Override the method to enforce that no Storage Connect classes be
        instantiated.
        """
        raise TypeError("Storage Connect classes cannot be instantiated")

    @staticmethod
    def __get_key_value_pairs__(key_val_s, result_dict = {}):
        """
        Create a dict from a string in the form of attr=value[,...].

        @param key_val_s: String containing key value pairs in key=val[,...] form.
        @type key_val_s: C{str}

        @param result_dict: Optional dictionary to update with the parsed string.
        @type result_dict: C{dict}

        @return: C{dict} with all the parsed key value pairs in the string.
        @rtype: C{dict}
        """
        for kv_vals in key_val_s.split(","):
            kv_pair = kv_vals.split("=")
            if len(kv_pair) == 2:
                result_dict[kv_pair[0]] = kv_pair[1]

            else:
                # Here we have a choice, either complain about the values
                # or just ignore it and run... I choose to complain since
                # I expect the user gave the values for a reason
                raise ValueFormatEx("Invalid format expecting an "
                                    "attr=value[,...] string",
                                    "extra_info")

        return result_dict

    @staticmethod
    def __set_key_value_pairs__(key_val_d):
        """
        This method will create a string in the form of key=value[,...] from a
        dict.

        @param key_val_d: C{dict} containing all the values to inserted in the
                          string.
        @type key_val_d: C{dict}

        @return: String in with all keys in the dict in the format of
                 key=value[,...]
        @rtype: C{str}
        """
        kv_str = ""
        for kv_key in key_val_d:
            kv_str = kv_str + "%s=%s," % (kv_key, key_val_d[kv_key])

        return kv_str[:-1]

    @staticmethod
    def copy(src, dst):
        """
        This will do a block copy from the source to the destination

        @param src: Source for the block copy operation.
        @type  src: C{str}

        @param dst: Destination for the block copy operation.
        @type  dst: C{str}
        """
        IPlugin.logger.debug("ENTER: copy()")
        p_cmd = ("/bin/dd "
                 "if=%s "
                 "of=%s "
                 "bs=1M "
                 "iflag=direct "
                 "oflag=direct "
                 "conv=nocreat,notrunc,sparse") % (src, dst)
        p_cmd = p_cmd.split(" ")
        dd_p = subprocess.Popen(p_cmd,
                                stdout = subprocess.PIPE,
                                stderr = subprocess.PIPE,
                                close_fds=True)
        (p_out, p_err) = dd_p.communicate()
        if dd_p.returncode != 0:
            raise IOError("Unable to copy %s to %s : %s" % (src, dst, p_err))

        IPlugin.logger.debug("EXIT : copy() - Success")

    @classmethod
    def getPluginType(klass):
        """
        Return the plugin type.

        @return: C{plugin_type} for this specific plugin
        @rtype:  C{str}
        """
        return klass.plugin_type

    @classmethod
    def getTemporaryFile(klass,
                         mode = 'w+b',
                         bufsize = -1,
                         suffix = '',
                         prefix = 'tmpfil',
                         directory = None):
        """
        This method is identical to the standard tempfile.TemporaryFile()
        method. Please see the tempfile.TemporaryFile() documentation on how to
        use it.
        """
        if directory == None:
            directory = klass.__plugin_temp_dir__
            if not os.path.exists(directory):
                os.makedirs(directory)

        if "tempfile" not in sys.modules:
            import tempfile

        new_tmp_file = tempfile.TemporaryFile(mode,
                                              bufsize,
                                              suffix,
                                              prefix,
                                              directory)
        if getattr(klass, "__temp_files__", None):
            klass.__temp_files__.append(new_tmp_file)

        else:
            klass.__temp_files__ = [new_tmp_file]

        return new_tmp_file

    @classmethod
    def getNamedTemporaryFile(klass,
                              mode = 'w+b',
                              bufsize = -1,
                              suffix = '',
                              prefix = 'tmpfil',
                              directory = None,
                              delete = True):
        """
        This method is identical to the standard tempfile.NamedTemporaryFile()
        method. Please see the tempfile.NamedTemporaryFile() documentation on
        how to use it.
        """
        if directory == None:
            directory = klass.__plugin_temp_dir__
            if not os.path.exists(directory):
                os.makedirs(directory)
        if "tempfile" not in sys.modules:
            import tempfile
        new_tmp_file = tempfile.NamedTemporaryFile(mode,
                                                   bufsize,
                                                   suffix,
                                                   prefix,
                                                   directory,
                                                   delete)
        if getattr(klass, "__temp_files__", None):
            klass.__temp_files__.append(new_tmp_file)
        else:
            klass.__temp_files__ = [new_tmp_file]
        return new_tmp_file

class IStorageArrayPlugin(IPlugin):
    """
    Plugin interface for Storage Array management.
    """

    name = "IStorageArrayPlugin"
    """
    @cvar: Class name.
    @type: C{str}
    """

    sa_api_version = ["1", "2", "7"]
    """
    @cvar: Storage Array Plugin API version.
    @type: C{list} of C{str}
    """

    storage_types = []
    """
    @ivar: Storage types this plugin can handle, see L{SANStorage} and
           L{iSCSIStorage}
    @type: C{list}
    """

    plugin_type = "isa"
    """
    @cvar: Oracle Storage Connect plugin type.
    @type: C{str}
    """

    __required_methods__ = {"validate":                     True,
                            "getCapabilities":              True,
                            "getStorageNames":              True,
                            "getInfo":                      True,
                            "getAccessGroups":              True,
                            "createAccessGroups":           True,
                            "removeAccessGroups":           True,
                            "addToAccessGroup":             True,
                            "renameAccessGroup":            True,
                            "removeFromAccessGroup":        True,
                            "discover":                     False,
                            "start":                        False,
                            "stop":                         False,
                            "refresh":                      False,
                            "list":                         True,
                            "updateSERecords":              False,
                            "getStatus":                    True,
                            "online":                       True,
                            "offline":                      True,
                            "create":                       True,
                            "startPresent":                 True,
                            "stopPresent":                  True,
                            "resize":                       True,
                            "remove":                       True,
                            "getCloneLimits":               True,
                            "isCloneable":                  True,
                            "clone":                        True,
                            "isSplittable":                 True,
                            "splitClone":                   True,
                            "cloneFromSnap":                True,
                            "getCurrentClones":             True,
                            "getSnapLimits":                True,
                            "isSnapable":                   True,
                            "createSnap":                   True,
                            "createMultiSnap":              True,
                            "isRestorable":                 True,
                            "snapRestore":                  True,
                            "snapRemove":                   True,
                            "getCurrentSnaps":              True,
                            "getQoSList":                   True,
                            "setQoS":                       True,
                            "getAsyncProgress":             False,
                            "deviceTeardown":               False,
                            "deviceResize":                 False,
                            "deviceSizeRefresh":            False,
                           }
    """
    @cvar: List of required methods for this L{Version<plugin_api_version>} of
           the API
    @type: C{dict}
    """

    @classmethod
    def getPluginInfo(klass):
        """
        Return all the required information to identify this plugin.

        @return: C{{"plugin_type",
                    "plugin_name",
                    "vendor_name",
                    "plugin_version",
                    "plugin_desc",
                    "storage_types",
                    "ss_extra_info_help",
                    "se_extra_info_help",
                    "required_api_vers",
                    "extended_api_version",
                    "sa_api_version",
                    "plugin_ability",
                    "generic_plugin"}}
        @rtype:  C{dict}
        """
        return {"plugin_type":          klass.plugin_type,
                "plugin_name":          klass.plugin_name,
                "vendor_name":          klass.vendor_name,
                "plugin_version":       klass.plugin_version,
                "plugin_desc":          klass.plugin_desc,
                "storage_types":        klass.storage_types,
                "ss_extra_info_help":   klass.ss_extra_info_help,
                "se_extra_info_help":   klass.se_extra_info_help,
                "required_api_vers":    klass.required_api_vers,
                "extended_api_version": klass.extended_api_version,
                "sa_api_version":       klass.sa_api_version,
                "plugin_ability":       klass.plugin_ability,
                "generic_plugin":       klass.__generic_plugin__,
               }

    @staticmethod
    def validate(ss_record):
        """
        Validate the Storage Server record.

        @param ss_record: L{Storage Server record<__SSRecord__>}.
        @type  ss_record: C{dict}.

        @raise IPluginException: on failure.
        """
        raise NotImplementedError()

    @staticmethod
    def getCapabilities(ss_record):
        """
        Obtain the capabilities from the Storage Array.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @return: L{Storage Server Capabilities<plugin_ability>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()


    @staticmethod
    def getStorageNames(ss_record):
        """
        Obtain a list of available storage names from which the user would
        choose the specific Storage Array to associate with this specific
        Storage Server record.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @return: list of available storage names OR if not required or supported
                 an empty list.
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getInfo(ss_record, se_record = None):
        """
        Obtain information on either the Storage Array or the Storage Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: Optional L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @return: updated L{Storage Server record<__SSRecord__>} OR
                 updated L{Storage Element record<__SERecord__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getAccessGroups(ss_record, se_record = None):
        """
        Get the access groups and access group entries for either the Storage
        Array or if specified the Storage Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: An existing presented L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @return: list of L{Access groups<__access_grp__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def createAccessGroups(ss_record, access_grps):
        """
        Create the access groups on the Storage Array.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param access_grps: List of L{Access Groups<__access_grp__>} to create
                            on the Storage Array.
        @type  access_grps: C{list}

        @return: list of L{Access groups<__access_grp__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def renameAccessGroup(ss_record, access_grp_name, new_access_grp_name):
        """
        Rename an access group on the Storage Array.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param access_grp_name: Name of the L{Access Group<__access_grp__>} to rename.
        @type  access_grp_name: C{str}

        @param new_access_grp_name: New name for the L{Access Group<__access_grp__>}.
        @type  new_access_grp_name: C{str}

        @return: list of L{Access groups<__access_grp__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def removeAccessGroups(ss_record, access_grps, forced_remove = False):
        """
        Remove the access groups from the Storage Array.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param access_grps: List of L{Access Groups<__access_grp__>} to remove
                            from the Storage Array.
        @type  access_grps: C{list}

        @param forced_remove: Flag if the access group should force fully be
                              removed from  the Storage Array.
        @type  forced_remove: C{bool}

        @return: list of L{Access groups<__access_grp__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def addToAccessGroup(ss_record, access_grp_name, grp_entries):
        """
        Get the access groups and access group entries for either the Storage
        Array or if specified the Storage Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param access_grp_name: Access group that the access control entries
                                should be added to.
        @type  access_grp_name: C{str}

        @param grp_entries: List of access control entries to add to the access
                            group.
        @type  grp_entries: C{list}

        @return: L{Access group<__access_grp__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def removeFromAccessGroup(ss_record, access_grp_name, grp_entries):
        """
        Get the access groups and access group entries for either the Storage
        Array or if specified the Storage Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param access_grp_name: Access group that the access control entries
                                should be removed from.
        @type  access_grp_name: C{str}

        @param grp_entries: List of access control entries to add to the access
                            group.
        @type  grp_entries: C{list}

        @return: L{Access group<__access_grp__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def discover(ss_record):
        """
        Discover the targets that the Storage Array makes available.

        @note: The is is only applicable to an iSCSI Storage Array

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @return: updated L{Storage Server record<__SSRecord__>} with the
                 storage_id field filled in.
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """


        IPlugin.logger.debug("ENTER: discover()")
        if ss_record["storage_type"] != IStorageArrayPlugin.iSCSIStorage:
            raise InvalidValueEx(_("Invalid storage type"))

        if ss_record.get("netdev", None):
            # TODO:
            # Set up the network device for the node session
            pass

        # leave the empty entry as the manager wants this
        ss_record["storage_id"] = [] 

        if ss_record.get("access_host", None):
            return _discover_access_host(ss_record)

        elif ss_record.get("access_descriptors", None):
            return _discover_access_descriptors(ss_record)

        else:
            raise InvalidValueEx(_("Invalid access fields"))

        IPlugin.logger.debug("EXIT : discover() - Success")
        return ss_record


    @staticmethod
    def start(ss_record):
        """
        Start access to this Storage Array.

        @note: The is is only applicable to an iSCSI Storage Array

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @raise IPluginException: on failure.
        """
        IPlugin.logger.debug("ENTER: start()")

        if ss_record["storage_type"] != IStorageArrayPlugin.iSCSIStorage:
            raise InvalidValueEx(_("Invalid storage type"))

        if ss_record.get("iscsi_storage_targets", None):

            if ss_record.get("access_host", None):
                san.iscsi_login(ss_record.get("username", None),
                                ss_record.get("passwd", None),
                                ss_record.get("chap", None),
                                ss_record["iscsi_storage_targets"])


            elif ss_record.get("access_descriptors", None):
                ss_ads = ss_record.get("access_descriptors", [])
                for ss_ad in ss_ads:
                    targets = _targets_access_descriptor(ss_record["iscsi_storage_targets"], ss_ad)
                    san.iscsi_login(ss_ad.get("username", None),
                                    ss_ad.get("passwd", None),
                                    ss_record.get("chap", None),
                                    targets)
        else:
            IPlugin.logger.warning("No targets to start")
        
        IPlugin.logger.debug("EXIT : start() - Success")



    @staticmethod
    def stop(ss_record):
        """
        Stop access to this Storage Array.

        @note: The is is only applicable to an iSCSI Storage Array

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @raise IPluginException: on failure.
        """
        IPlugin.logger.debug("ENTER: stop()")

        if ss_record["storage_type"] != IStorageArrayPlugin.iSCSIStorage:
            raise InvalidValueEx(_("Invalid storage type"))

        if not ss_record.get("iscsi_storage_targets", None):
            IPlugin.logger.warning("No targets to stop")
            return

        san.iscsi_logout(ss_record["iscsi_storage_targets"])
        
        IPlugin.logger.debug("EXIT : stop() - Success")


    @staticmethod
    def refresh(ss_record):
        """
        Initiate a rescan for newly added or removed LUNs.

        @note: This does not return a list of the new or or removed LUNs, it
               just initiate a rescan.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @raise IPluginException: on failure.
        """
        IPlugin.logger.debug("ENTER: refresh()")
        if ss_record["storage_type"] == IStorageArrayPlugin.iSCSIStorage:
            san.iscsi_rescan(ss_record)


        elif ss_record["storage_type"] == IStorageArrayPlugin.SANStorage:
            san.fc_refresh()

        else:
            raise InvalidValueEx(_("Invalid storage type"))

        san.mp_refresh()

        IPlugin.logger.debug("EXIT : refresh() - Success")


    @staticmethod
    def deviceTeardown(ss_record, se_record):
        """
        Tear down multipath device and associated slaves.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: Optional L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @raise IPluginException: on failure.
        """
        IPlugin.logger.debug("ENTER: deviceTearDown()")
        san.mp_destroy(se_record["page83_id"])
        IPlugin.logger.debug("EXIT : deviceTearDown() - Success")


    @staticmethod
    def deviceSizeRefresh(ss_record, se_record):
        """
        Refresh the size of a device.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: Optional L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @raise IPluginException: on failure.
        """
        IPlugin.logger.debug("ENTER: deviceSizeRefresh()")
        san.device_resize(ss_record, se_record["page83_id"])
        IPlugin.logger.debug("EXIT : deviceSizeRefresh() - Success")


    @staticmethod
    def list(ss_record, se_type = [IPlugin.LUNType,
                                   IPlugin.SnapCloneType]):
        """
        Return the list of Storage Entities associated with the Storage
        Array.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_type: List of Storage Element types to be returned.
        @type  se_type: C{list}

        @return: list of L{Storage Element records<__SERecord__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def updateSERecords(ss_record, se_records, append_allowed = False):
        """
        Update the list of Storage Element records. The method should
        complete the each of the Storage Element records, including the
        device path if the Storage Element is visible on the server where
        the method is executing. NOTE: The plug-in should use the
        L{dev_path_prefix} value to determine the device path (if any) for
        each of the Storage Elements.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type ss_record: C{dict}

        @param se_records: List of Storage Element records to examine and
                           update with all the required information, this
                           includes the local device path, if it is
                           available on the server.
        @type  se_records: C{list}

        @param append_allowed: By default the method is not allowed to add
                               new Storage Element records to the list
                               passed in, if this flag is set the method is
                               allowed to add Storage Element records to the
                               list for newly discovered Storage Elements.
        @type  append_allowed: C{bool}

        @return: list of updated L{Storage Element records<__SERecord__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """

        IPlugin.logger.debug("ENTER: updateSERecord()")

        if len(se_records) < 1 :
            IPlugin.logger.debug("EXIT : updateSERecord() - Nothing to do!")
            return se_records

        for se_record in se_records:
            if not se_record.has_key('page83_id'):
                IPlugin.logger.error("SERecord does not contain a page83_id")
                continue

            dev_name = san.device_path(se_record["page83_id"])
            if dev_name:
                se_record["path"].append(dev_name)
            else:
                IPlugin.logger.error("Device %s does not exist",
                                     se_record["page83_id"])

        IPlugin.logger.debug("EXIT : updateSERecord() - Success")
        return se_records


    @staticmethod
    def getStatus(ss_record, se_record = None):
        """
        Return the status of the Storage Array or the Storage Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: Optional L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @return: Status of the Storage Array or the Storage Element
        @rtype:  C{str}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def online(ss_record, se_record):
        """
        Online the Storage Element on the Storage Array (if offline).

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def offline(ss_record, se_record):
        """
        Offline the Storage Element on the Storage Array (if offline).

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def create(ss_record, se_record, thin_provision = True, qos = None):
        """
        Provision a new Storage Element on the Storage Array.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @param thin_provision: Indicate if the Storage Element should be thin
                               provisioned (if the Storage Array support thin
                               provisioning) or fully provisioned.
        @type  thin_provision: C{bool}

        @param qos: Desired Quality-of-Service for the new Storage Element
        @type  qos: C{str}

        @return: updated L{Storage Element record<__SERecord__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def startPresent(ss_record, se_record, access_grp_names):
        """
        Present a Storage Element on Storage Array to the list of ports.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @param access_grp_names: Names of the access groups the Storage Element
                                 should be presented to.
        @type  access_grp_names: C{list}

        @return: updated L{Storage Element record<__SERecord__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def stopPresent(ss_record, se_record, access_grp_names):
        """
        Unpresent a Storage Element on Storage Array to the list of ports.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @param access_grp_names: Names of the access groups the Storage Element
                                 should stop being be presented to.
        @type  access_grp_names: C{list}

        @return: updated L{Storage Element record<__SERecord__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def resize(ss_record, se_record, new_size):
        """
        Resize the Storage Element on the Storage Array.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @param new_size: New desired size for the Storage Element.
        @type  new_size: C{int}

        @return: updated L{Storage Element record<__SERecord__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def remove(ss_record, se_record):
        """
        Remove a Storage Element from the Storage Array.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getCloneLimits(ss_record, se_record = None):
        """
        Return the maximum number of shallow (thin) clones allowed for a
        Storage Array or Storage Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: Optional L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @return: Maximum number (-1 for unlimited) of shallow clones allowed
                 for the Storage Array or the specific type of Storage Element
                 if supplied.
        @rtype:  C{int}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def isCloneable(ss_record, se_record):
        """
        Check if shallow (thin) clone creations is allowed for the Storage
        Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @return: True if a shallow clone can be created for the Storage Element.
        @rtype:  C{bool}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def clone(ss_record, se_record, dest_se_record = None, qos = None):
        """
        Create a shallow (thin) clone of the Storage Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @param dest_se_record: Destination L{Storage Element record<__SERecord__>}
        @type  dest_se_record: C{dict}

        @param qos: Desired Quality-of_service for the new clone
        @type  qos: C{str}

        @return: Newly created L{Storage Element record<__SERecord__>} for the
                 cloned Storage Element.
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def isSplittable(ss_record, se_record):
        """
        Determine if a deep clone can be made from a shallow (thin) clone (split the clones).

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @return: True if the clones can be split.
        @rtype:  C{bool}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def splitClone(ss_record, se_record):
        """
        Create a deep clone from a shallow (thin) clone (split the clones).

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def cloneFromSnap(ss_record, se_record, snap_se_record,
                      dest_se_record = None, qos = None):
        """
        Create a shallow (thin) clone using a snapshot.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: Original L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @param snap_se_record: Snapshot L{Storage Element record<__SERecord__>}
                               to be clone from
        @type  snap_se_record: C{dict}

        @param dest_se_record: Destination L{Storage Element record<__SERecord__>}
        @type  dest_se_record: C{dict}

        @param qos: Desired Quality-of_service for the new clone
        @type  qos: C{str}

        @return: created L{Storage Element record<__SERecord__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getCurrentClones(ss_record, se_record = None):
        """
        Return a list of the current clones for a Storage Array or Storage
        Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: Optional L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @return: List of current known clones on the Storage Array or Storage
                 Element if supplied.
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getSnapLimits(ss_record, se_record = None):
        """
        Return the maximum number of snapshots allowed for a Storage Array or
        Storage Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: Optional L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @return: Maximum number (-1 for unlimited) of snapshots allowed for the
                 Storage Array or the specific type of Storage Element if
                 supplied.
        @rtype:  C{int}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def isSnapable(ss_record, se_record = None):
        """
        Check if a snapshots can be created on the Storage Array or if
        specified for the specific Storage Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: Optional L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @return: True if a snapshots can be created on the Storage Array or if
                 specified for the specific Storage Element.
        @rtype:  C{bool}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def createSnap(ss_record, se_record, snap_se_record = None):
        """
        Create a snapshot of the Storage Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @param snap_se_record: Optional L{Storage Element record<__SERecord__>}
                               to specify the snapshot name and destination
        @type  snap_se_record: C{dict}

        @return: created L{Storage Element record<__SERecord__>} for the snapshot
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def createMultiSnap(ss_record, se_records, snap_se_record = None):
        """
        Create a snapshot of all the listed Storage Entities together.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record:  C{dict}

        @param se_records: L{Storage Element record<__SERecord__>}
        @type  se_records:  C{dict}

        @param snap_se_record: Optional L{Storage Element record<__SERecord__>}
                               to specify the snapshot name and destination
        @type  snap_se_record:  C{dict}

        @return: Created L{Storage Element record<__SERecord__>} for the
                 snapshot.
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def isRestorable(ss_record, snap_se_record, se_record = None):
        """
        Check if the Storage Array or Storage Element can be rolled back to the
        named snapshot.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param snap_se_record: L{Storage Element record<__SERecord__>} naming
                               the snapshot to be rolled back to.
        @type  snap_se_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @return: True if a snapshot can be rolled back to the snapshot,
                 otherwise return False.
        @rtype:  C{bool}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def snapRestore(ss_record, snap_se_record, se_record = None):
        """
        Roll the Storage Array or Storage Element back to the named snapshot.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param snap_se_record: L{Storage Element record<__SERecord__>} naming
                               the snapshot to be rolled back to.
        @type  snap_se_record: C{dict}

        @param se_record: L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def snapRemove(ss_record, snap_se_record):
        """
        Destroy the named snapshot.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param snap_se_record: L{Storage Element record<__SERecord__>}
        @type  snap_se_record: C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getCurrentSnaps(ss_record, se_record = None):
        """
        Return a list of the current snapshots for a Storage Array or Storage
        Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: Optional L{Storage Element record<__SERecord__>}
        @type  se_record: C{dict}

        @return: List of L{Storage Element records<__SERecord__>} for all known
                 snapshots on the Storage Array or Storage Element if supplied.
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getQoSList(ss_record):
        """
        Return a list of possible QoS values for the Storage Array.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @return: List of L{qos_vals<__qos_vals__>} dicts for all the possible
                 QoS values on the Storage Array.
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def setQoS(ss_record, se_record, qos):
        """
        Set the QoS for the Storage Element to the QoS supplied.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param se_record: L{Storage Server record<__SSRecord__>}
        @type  se_record: C{dict}

        @param qos: Desired Quality-of-Service for the new clone.
        @type  qos: C{str}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getAsyncProgress(ss_record, some_record):
        """
        Obtain Async Progress. This should never be called as
        the plugin should overide this method if it is in
        asynch mode.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param some_record: Record previously returned from the
                            call that started the asynchronous
                            operation with the async_progress and
                            if required the async_handle fields
                            added (and not set to None).
        @type  some_record: C{dict}

        @return: Fully completed record that would have been
                 returned if the initial call completed
                 synchronously except it have the async_progress field.
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """

        raise NotImplementedError()

class IFileSystemPlugin(IPlugin):
    """
    Plugin interface for file system management.
    """

    name = "IFileSystemPlugin"
    """
    @cvar: Class name.
    @type: C{str}
    """

    fs_api_version = ["1", "2", "7"]
    """
    @cvar: File system plugin API version.
    @type: C{list} of C{str}
    """

    filesys_type = ""
    """
    @cvar: File system type this plugin can handle, either L{LocalFileSystem}
           or L{NetworkFileSystem}.
    @type: C{str}
    """

    filesys_name = ""
    """
    @cvar: Name of the file system to show to the user when selecting which file
           system to use, for example "OCFS2" or "ZFS" etc.
    @type: C{str}
    """

    plugin_type = "ifs"
    """
    @cvar: Oracle Storage Connect plugin type.
    @type: C{str}
    """

    __required_methods__ = {"validate":                     True,
                            "getCapabilities":              True,
                            "getStorageNames":              True,
                            "getStorageServerInfo":         True,
                            "getFileSystemInfo":            True,
                            "getFileInfo":                  True,
                            "getAccessGroups":              True,
                            "createAccessGroups":           True,
                            "renameAccessGroup":            True,
                            "removeAccessGroups":           True,
                            "addToAccessGroup":             True,
                            "removeFromAccessGroup":        True,
                            "listFileSystems":              True,
                            "listMountPoints":              True,
                            "list":                         True,
                            "getStatus":                    True,
                            "mount":                        True,
                            "unmount":                      True,
                            "createFileSystem":             True,
                            "create":                       True,
                            "startPresent":                 True,
                            "stopPresent":                  True,
                            "resizeFileSystem":             True,
                            "resize":                       True,
                            "destroyFileSystem":            True,
                            "destroy":                      True,
                            "getFileSystemCloneLimits":     True,
                            "getCloneLimits":               True,
                            "isCloneable":                  True,
                            "clone":                        False,
                            "isSplittable":                 True,
                            "splitClone":                   True,
                            "cloneFromSnap":                True,
                            "getCurrentClones":             True,
                            "getFileSystemSnapLimits":      True,
                            "getSnapLimits":                True,
                            "isSnapable":                   True,
                            "createSnap":                   True,
                            "createMultiSnap":              True,
                            "isRestorable":                 True,
                            "snapRestore":                  True,
                            "snapRemove":                   True,
                            "getCurrentSnaps":              True,
                            "getQoSList":                   True,
                            "setQoS":                       True,
                            "getAsyncProgress":             False,
                           }
    """
    @cvar: List of required methods for this L{Version<plugin_api_version>} of
           the API
    @type: C{dict}
    """

    @classmethod
    def getPluginInfo(klass):
        """
        Return all the required information to identify this plugin

        @return: C{{"plugin_type",
                    "plugin_name",
                    "vendor_name",
                    "plugin_version",
                    "plugin_desc",
                    "filesys_type",
                    "filesys_name",
                    "ss_extra_info_help",
                    "fs_extra_info_help",
                    "file_extra_info_help",
                    "required_api_vers",
                    "extended_api_version",
                    "fs_api_version",
                    "plugin_ability",
                    "generic_plugin"}}
        @rtype:  C{dict}
        """
        return {"plugin_type":          klass.plugin_type,
                "plugin_name":          klass.plugin_name,
                "vendor_name":          klass.vendor_name,
                "plugin_version":       klass.plugin_version,
                "plugin_desc":          klass.plugin_desc,
                "filesys_type":         klass.filesys_type,
                "filesys_name":         klass.filesys_name,
                "ss_extra_info_help":   klass.ss_extra_info_help,
                "fs_extra_info_help":   klass.fs_extra_info_help,
                "file_extra_info_help": klass.file_extra_info_help,
                "required_api_vers":    klass.required_api_vers,
                "extended_api_version":  klass.extended_api_version,
                "fs_api_version":       klass.fs_api_version,
                "plugin_ability":       klass.plugin_ability,
                "generic_plugin":       klass.__generic_plugin__,
                "cluster_required":     klass.__cluster_required__,
               }

    @staticmethod
    def validate(ss_record):
        """
        Validate the Storage Server record.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getCapabilities(ss_record):
        """
        Obtain the capabilities from the Storage Server.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @return: L{Storage Server Capabilities<plugin_ability>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()


    @staticmethod
    def getStorageNames(ss_record):
        """
        Obtain a list of available storage names from which the user would
        choose the specific Storage Server to associate with this specific
        Storage Server record.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @return: list of available storage names OR if not required or supported
                 an empty list.
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getStorageServerInfo(ss_record):
        """
        Obtain information on the file File Server.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @return: updated L{Storage Server record<__SSRecord__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getFileSystemInfo(ss_record, fs_record):
        """
        Obtain information on either the File Server or the File System.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @return: updated L{File System record<__FSRecord__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getFileInfo(ss_record, fs_record, file_record, file_comp_record = None):
        """
        Obtain information on the file or if two file records are given, info
        on both files including how much (if any) data blocks are shared
        by the files.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @param file_comp_record: Optional L{File record<__FileRecord__>}
        @type  file_comp_record: C{dict}

        @return: updated L{File record<__FileRecord__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getAccessGroups(ss_record, fs_record = None):
        """
        Get the access groups and access group entries for either the Storage
        Server or if specified the File System.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: Optional L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @return: list of L{Access groups<__access_grp__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def createAccessGroups(ss_record, access_grps):
        """
        Create the access groups on the Storage Server.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param access_grps: List of L{Access Groups<__access_grp__>} to create
                            on the Storage Server.
        @type  access_grps: C{list}

        @return: list of L{Access groups<__access_grp__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def renameAccessGroup(ss_record, access_grp_name, new_access_grp_name):
        """
        Rename an access group on the Storage Server.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param access_grp_name: Name of the L{Access Group<__access_grp__>} to rename.
        @type  access_grp_name: C{str}

        @param new_access_grp_name: New name for the L{Access Group<__access_grp__>}.
        @type  new_access_grp_name: C{str}

        @return: list of L{Access groups<__access_grp__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def removeAccessGroups(ss_record, access_grps):
        """
        Remove the access groups from the Storage Server.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param access_grps: List of L{Access Groups<__access_grp__>} to remove
                            from the Storage Server.
        @type  access_grps: C{list}

        @return: list of L{Access groups<__access_grp__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def addToAccessGroup(ss_record, access_grp_name, grp_entries):
        """
        Get the access groups and access group entries for either the Storage
        Server or if specified the File System.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param access_grp_name: Access group that the access control entries
                                should be added to.
        @type  access_grp_name: C{str}

        @param grp_entries: List of access control entries to add to the access
                            group.
        @type  grp_entries: C{list}

        @return: L{Access group<__access_grp__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def removeFromAccessGroup(ss_record, access_grp_name, grp_entries):
        """
        Get the access groups and access group entries for either the Storage
        Server or if specified the File System.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param access_grp_name: Access group that the access control entries
                                should be removed from.
        @type  access_grp_name: C{str}

        @param grp_entries: List of access control entries to add to the access
                            group.
        @type  grp_entries: C{list}

        @return: L{Access group<__access_grp__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def listFileSystems(ss_record, include_unpresented = False):
        """
        Return the list of available (or known) File Systems.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param include_unpresented: True if ALL File Systems on the File Server
                                    should be returned not just the available ones.
        @type  include_unpresented: C{bool}

        @return: list of L{File System records<__FSRecord__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def listMountPoints(ss_record, fs_record = None):
        """
        Return the list of mount points the plugin can manage or if the FS Record
        is supplied, all the mount points for the specific file system.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: Optional L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @return: list of L{Moint point records<__MountRecord__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def list(ss_record, fs_record, file_record, recursive = False):
        """
        Return the list of files.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @param recursive: True if the list should traverse all directories
                          below the directory specified in the
                          L{file_record<__FileRecord__>}.
        @type  recursive: C{bool}

        @return: list of L{File records<__FileRecord__>}
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getStatus(ss_record, fs_record = None):
        """
        Return the status of the File System.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: Optional L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @return: Status of the File System or File Server
        @rtype:  C{str}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def mount(ss_record, fs_record, mount_point, share_path = None, create_mount_point = False, mount_options = None):
        """
        Mount the File System if not already mounted.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param mount_point: Directory where File System should be mounted.
        @type  mount_point: C{str}

        @param share_path: Subdirectory of File System could be mounted.
        @type  share_path: C{str}

        @param create_mount_point: Optional flag to indicate that the mount point should be created.
        @type  create_mount_point: C{bool}

        @param mount_options: Optional list of mount options
        @type  mount_options: C{list}

        @return: L{File System record<__FSRecord__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def unmount(ss_record, fs_record, mount_point, destroy_mount_point = False):
        """
        Unmount the File System if mounted.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param mount_point: Directory of the File System that should be un-mounted.
        @type  mount_point: C{str}

        @param destroy_mount_point: Optional flag to indicate that the mount point should be removed.
        @type  destroy_mount_point: C{bool}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def createFileSystem(ss_record,
                         name,
                         backing_device,
                         size = 0,
                         access_grp_names = None,
                         qos = None,
                         force = False):
        """
        Create a new File System.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param name: User friendly name for the new File System
        @type  name: C{str}

        @param backing_device: File System backing device, can
                               be either a string with the backing
                               device if only a single device is
                               supported (as indicated in the
                               plugin ability dict) or a list
                               of backing devices.
        @type  backing_device: C{str} or C{list}

        @param size: Desired size for the new File System (0 implies as large
                     as possible)
        @type  size: C{int}

        @param access_grp_names: Desired Quality-of-Service for the new File
                                 System
        @type  access_grp_names: C{list}

        @param qos: Desired Quality-of-Service for the new File System
        @type  qos: C{str}

        @param force: Flag to indicate any existing File System should be
                      overwritten
        @type  force: C{bool}

        @return: new L{File System record<__FSRecord__>}
        @rtype:  C{dict}

        @note: ONLY overwrite an existing File System if the force option is
               specified.

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def create(ss_record, fs_record, file_name, file_type, size, sparse = True, force = False):
        """
        Create a new File.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_name: Name of the file to create.
        @type  file_name: C{str}

        @param file_type: Type of file to create. will be one of
                          L{Known File Types<__FRTypes__>}
        @type  file_type: C{str}

        @param size: Size of the file to create.
        @type  size: C{int}

        @param sparse: Flag to indicate the File should be sparsely allocated (if possible).
        @type  sparse: C{bool}

        @param force: Flag to indicate any existing File should be overwritten.
        @type  force: C{bool}

        @return: updated L{File record<__FileRecord__>}
        @rtype:  C{dict}

        @note: ONLY overwrite an existing File if the force option is specified.

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def startPresent(ss_record, fs_record, access_grp_names):
        """
        Present a File system on the File Server to the list of nodes.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param access_grp_names: Names of the access groups the File System
                                 should be presented to.
        @type  access_grp_names: C{list}

        @return: updated L{File System record<__FSRecord__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def stopPresent(ss_record, fs_record, access_grp_names):
        """
        Unpresent a File System on the File Server to the list of nodes.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param access_grp_names: Names of the access groups the File System
                                 should stop being be presented to.
        @type  access_grp_names: C{list}

        @return: updated L{File System record<__FSRecord__>}
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def resizeFileSystem(ss_record, fs_record, new_size):
        """
        Resize the File System.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param new_size: New desired size for the File System.
        @type  new_size: C{int}

        @return: updated L{File System record<__FSRecord__>}
        @rtype:  C{dict}

        @note: It is the plugin responsibilty to verify that all conditions are
               met for the resize operation. For example if online resize is not
               supported, the plugin must make sure the file system is not
               mounted and raise an Exception if it is.

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def resize(ss_record, fs_record, file_record, new_size, sparse = True):
        """
        Resize the File.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @param new_size: New desired size for the file.
        @type  new_size: C{int}

        @param sparse: Flag to indicate the File should be extended using sparse
                       (if possible).
        @type  sparse: C{bool}

        @return: updated L{File record<__FileRecord__>}
        @rtype:  C{dict}

        @note: It is the plugin responsibilty to verify that all conditions are
               met for the resize operation. For example if online resize is not
               supported, the plugin must make sure the file is not currently
               open and/or locked and raise an Exception if it is.

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def destroyFileSystem(ss_record, fs_record):
        """
        Remove a File System from the File Server.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def destroy(ss_record, fs_record, file_record):
        """
        Remove a file from the File System on the File Server.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getFileSystemCloneLimits(ss_record, fs_record = None):
        """
        Return the maximum number of clones allowed for a File Server or
        File System.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: Optional L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @return: Maximum number (-1 for unlimited) of clones allowed for the
                 File Server or the specific file system if supplied.
        @rtype:  C{int}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getCloneLimits(ss_record, fs_record, file_record):
        """
        Return the maximum number of snapshots allowed for a File Server or
        File System.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @return: Maximum number (-1 for unlimited) of clones allowed for the
                 File.
        @rtype:  C{int}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def isCloneable(ss_record, fs_record, file_record):
        """
        Check if shallow (thin) clone creations is allowed for the file on the
        File System.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @return: True if a snapshot can be created for the File record.
        @rtype:  C{bool}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def clone(ss_record, fs_record, file_record, dest_file_name = None):
        """
        Create a shallow clone of the file on the file system.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @param dest_file_name: Optional destination file name
        @type  dest_file_name: C{str}

        @return: newly created L{File record<__FileRecord__>} for the
                 cloned storage element
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def isSplittable(ss_record, fs_record, file_record):
        """
        Determine if a deep clone can be made from a shallow (thin) clone (split the clones).

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @return: True if the clones can be split.
        @rtype:  C{bool}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def splitClone(ss_record, fs_record, file_record):
        """
        Create a deep clone from a shallow clone (Split the clones).

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def cloneFromSnap(ss_record, fs_record, snap_file_record, dest_file_name = None):
        """
        Create a shallow clone from a named snapshot.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param snap_file_record: L{File record<__FileRecord__>}
        @type  snap_file_record: C{dict}

        @param dest_file_name: Optional destination file name
        @type  dest_file_name: C{str}

        @return: created L{File record<__FileRecord__>} for the new clone
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getCurrentClones(ss_record, fs_record, file_record = None):
        """
        Return a list of the current clones for the File System or Storage
        Element.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: Optional L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @return: List of current known clones on the L{File record<__FileRecord__>}.
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getFileSystemSnapLimits(ss_record, fs_record = None):
        """
        Return the maximum number of snapshots allowed for a Storage Array or
        File System.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: Optional L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @return: Maximum number (-1 for unlimited) of snapshots allowed for the
                 Storage Array or the specific file system if supplied.
        @rtype:  C{int}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getSnapLimits(ss_record, fs_record, file_record):
        """
        Return the maximum number of snapshots allowed for a specific file on
        the file system.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @return: Maximum number (-1 for unlimited) of snapshots allowed for the
                specific file.
        @rtype:  C{int}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def isSnapable(ss_record, fs_record, file_record):
        """
        Check if a snapshot creation is allowed for the file on the file system
        file.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @return: True if a snapshot can be created for the File record.
        @rtype:  C{bool}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def createSnap(ss_record, fs_record, file_record, snap_name = None):
        """
        Create a snapshot of the file for the file system.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @param snap_name: Optional snapshot name
        @type  snap_name: C{str}

        @return: created L{File record<__FileRecord__>} for the snapshot
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def createMultiSnap(ss_record, fs_record, file_records, snap_name = None):
        """
        Create a snapshot of all the files listed for the file system.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_records: List of L{File record<__FileRecord__>}
        @type  file_records: C{list}

        @param snap_name: Optional snapshot name
        @type  snap_name: C{str}

        @return: Created L{File record<__FileRecord__>} for the snapshot
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def isRestorable(ss_record, fs_record, file_record, snap_file_record):
        """
        Check if the Storage Array or Storage Element can be rolled back to the
        named snapshot.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{Storage Element record<__SERecord__>} naming
                            the file tjat should be rolled back.
        @type  file_record: C{dict}

        @param snap_file_record: L{Storage Element record<__SERecord__>} naming
                                 the snapshot to be rolled back to.
        @type  snap_file_record: C{dict}

        @return: True if a snapshot can be rolled .
        @rtype:  C{bool}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def snapRestore(ss_record, fs_record, file_record, snap_file_record):
        """
        Roll the file back to the named snapshot.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: L{File record<__FileRecord__>} of the file to roll
                            back.
        @type  file_record: C{dict}

        @param snap_file_record: L{File record<__FileRecord__>} naming the
                                 snapshot to roll back to.
        @type  snap_file_record: C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def snapRemove(ss_record, fs_record, snap_file_record):
        """
        Destroy the named snapshot.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param snap_file_record: L{File record<__FileRecord__>}
        @type  snap_file_record: C{dict}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getCurrentSnaps(ss_record, fs_record, file_record = None):
        """
        Return a list of the current snapshots for a File System or File Record.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param file_record: Optional L{File record<__FileRecord__>}
        @type  file_record: C{dict}

        @return: List of current known snapshots for the File System or File
                 record if supplied.
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getQoSList(ss_record, fs_record = None):
        """
        Return a list of possible QoS values for the File System.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: Optional L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @return: List of L{qos_vals<__qos_vals__>} dicts for all the possible
                 QoS values for either the Storage Server or the File System.
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def setQoS(ss_record, fs_record, qos):
        """
        Set the QoS for the Storage Element to the QoS supplied.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @param qos: Desired Quality-of-Service for the new clone.
        @type  qos: C{str}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getFileSystemBackingDevices(ss_record, fs_record):
        """
        Get the list of valid backing devices that can be used to create
        a file system on.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param fs_record: L{File System record<__FSRecord__>}
        @type  fs_record: C{dict}

        @return: List of L{backing_device<__BackingDeviceRecord__>} dicts
                 for all the possible backing devices that can be used to
                 create a new file system on.
        @rtype:  C{list}

        @raise IPluginException: on failure
        """
        raise NotImplementedError()

    @staticmethod
    def getAsyncProgress(ss_record, some_record):
        """
        Obtain Async Progress. This should never be called as
        the plugin should overide this method if it is in
        asynch mode.

        @param ss_record: L{Storage Server record<__SSRecord__>}
        @type  ss_record: C{dict}

        @param some_record: Record previously returned from the
                            call that started the asynchronous
                            operation with the async_progress and
                            if required the async_handle fields
                            added (and not set to None).
        @type  some_record: C{dict}

        @return: Fully completed record that would have been
                 returned if the initial call completed
                 synchronously except it have the async_progress field.
        @rtype:  C{dict}

        @raise IPluginException: on failure
        """

        raise NotImplementedError()

class IPluginException(Exception):
    """
    All IPlugin exceptions inherits from IPluginException to allow for blanket
    exception catching.
    """
    def __init__(self, msg):
        Exception.__init__(self, msg)
        self.msg = msg

    def __str__(self):
        return repr(self.msg)

class NoSuchOperationEx(IPluginException):
    """
    Exception that gets raised if the call is not valid for the plugin.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class OperationFailedEx(IPluginException):
    """
    Exception that gets raised if the operation failed.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class MissingKeyEx(IPluginException):
    """
    Exception that gets raised if a required key is not in the record.
    """
    def __init__(self, msg, key_name):
        IPluginException.__init__(self, msg)
        self.key_name = key_name

class ValueFormatEx(IPluginException):
    """
    Exception that gets raised when the plugin is unable to interpret the value
    passed in a key.
    """
    def __init__(self, msg, key_name):
        IPluginException.__init__(self, msg)
        self.key_name = key_name

class StorageArrayLicenseEx(IPluginException):
    """
    Exception that gets raised if the Storage Array feature is not licensed.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class TargetDiscoveryEx(IPluginException):
    """
    Exception that gets raised if discovery failed.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class LoginFailedEx(IPluginException):
    """
    Exception that gets raised if the login failed.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class LogoutFailedEx(IPluginException):
    """
    Exception that gets raised if the logout failed.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class RefreshFailedEx(IPluginException):
    """
    Exception that gets raised if the storage array refresh failed.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class ListFailedEx(IPluginException):
    """
    Exception that gets raised if the storage element list failed.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class CreateSnapFailedEx(IPluginException):
    """
    Exception that gets raised if a snapshot creation has failed.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class ListSnapFailedEx(IPluginException):
    """
    Exception that gets raised if listing snapshots failed.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class SnapRestoreNotSafeEx(IPluginException):
    """
    Exception that gets raised if snapshot cannot safely be rolled back.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class CloneFailedEx(IPluginException):
    """
    Exception that gets raised if a clone opertaion failed.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class InvalidStorageArrayEx(IPluginException):
    """
    Exception that gets raised if the ss_record is invalid.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class InvalidValueEx(IPluginException):
    """
    Exception that gets raised on an invalid value in one of the records.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class PermissionDeniedEx(IPluginException):
    """
    Exception that gets raised if permission is denied for the operation.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class StorageElementBusyEx(IPluginException):
    """
    Exception that gets raised when the operation is not allowed while the
    Storage Element is still in use.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class OperationPreReqNotMetEx(IPluginException):
    """
    Exception that gets raised if a operation's prerequisite have not been met.
    """
    def __init__(self, msg, pre_req):
        IPluginException.__init__(self, msg)
        self.pre_req = pre_req

class StorageNameRequiredEx(IPluginException):
    """
    Exception that gets raised by the validate() method if the plug-in require
    a storage name to be set in the ss_record.
    """
    def __init__(self, msg, pre_req):
        IPluginException.__init__(self, msg)
        self.pre_req = pre_req

class InvalidFSTypeEx(IPluginException):
    """
    Exception that gets raised if the FS record file system type is invalid.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class FileSystemBusyEx(IPluginException):
    """
    Exception that gets raised if the FS is busy.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class CommandMissingEx(IPluginException):
    """
    Exception that gets raised if the comand is not available.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

class FileSystemAlreadyMountedEx(IPluginException):
    """
    Exception that gets raised if the FS is already mounted.
    """
    def __init__(self, msg):
        IPluginException.__init__(self, msg)

def _discover_access_host(ss_record):       

    username = ss_record.get("username", None)
    passwd = ss_record.get("passwd", None)
    chap = ss_record.get("chap", None)
    
    if not ss_record.get("access_port", None):
        ss_record["access_port"] = 3260

    ss_record["iscsi_storage_targets"] = san.iscsi_discovery(
        ss_record["access_host"], ss_record["access_port"],
        username, passwd, chap)

    IPlugin.logger.debug("Discovered targets access host %s: " % ss_record["iscsi_storage_targets"])

    return ss_record

    
def _discover_access_descriptors(ss_record):       

    ss_ads = ss_record.get("access_descriptors", [])
    ss_record["iscsi_storage_targets"] = []

    for ss_ad in ss_ads:
        if not ss_ad.get("access_port", None):
            ss_ad["access_port"] = 3260

        username = ss_ad.get("username", None)
        passwd = ss_ad.get("passwd", None) 
        chap = ss_record.get("chap", None)

        targets = san.iscsi_discovery(ss_ad["access_host"],
                                      ss_ad["access_port"],
                                     username, passwd, chap)
        if targets:
            ss_record["iscsi_storage_targets"].extend(targets)

    IPlugin.logger.debug("Discovered targets access desc %s: " % ss_record["iscsi_storage_targets"])

    return ss_record


def _targets_access_descriptor(target_list, ss_ad):
 
    # target_list = [{'iscsi_target_name': 'iqn.1992-08.com:sn.101184323', 'iscsi_portal_address': '10.196.210.26', 'iscsi_portal_port': 3260} ...]
    portal = "%s:%s" % (socket.gethostbyname(ss_ad["access_host"]),
                                             str(ss_ad["access_port"]))
    targets = []
 
    for target in target_list:
        target_portal = "%s:%s" % (target['iscsi_portal_address'], target['iscsi_portal_port'])
        if target_portal == portal:
            targets.append(target)

    return (targets)
 


