#!/usr/bin/env python

"""
@copyright: Copyright (C) 2013, 2016 Oracle and/or its affiliates. All rights reserved.

@license:   See the file COPYING for redistribution information.

@summary:   Linux specific functions.

"""

import gettext
import os
import subprocess
import re
import socket
import time

import OSCPluginUtils


_ = gettext.gettext


def fc_refresh():
    """
    Initiate a rescan for newly added or removed FC LUNs.

    """
    from OSCPlugin import IPlugin
    from OSCPlugin import RefreshFailedEx
    
    if not os.path.exists("/sys/class/fc_host"):
        return

    for fc_host in os.listdir("/sys/class/fc_host"):
        fc_path = "/sys/class/fc_host/%s" % fc_host
        try:
            if os.path.exists("%s/port_state" % fc_path):
                status_f = open("%s/port_state" % fc_path, "r", 0)
                host_state = status_f.read().strip().upper()
                status_f.close()
                if host_state not in ["ONLINE",
                                      "ACTIVE",
                                      "LINK",
                                      "LINKUP",
                                      ]:
                    IPlugin.logger.debug("FC %s not Online - Skipping" % fc_host)
                    continue

        except IOError, ioe:
            IPlugin.logger.info("IO Error (%s) for FC %s - Skipping" % (ioe.strerror, fc_host))
            continue

        try:
            if os.path.exists("%s/issue_lip" % fc_path):
                lip_f = open("%s/issue_lip" % fc_path, "w", 0)
                lip_f.write("1")
                lip_f.close()
                time.sleep(0.0531)

                lipWait = 600
                while lipWait > 0:
                    try:
                        status_f = open("%s/port_state" % fc_path, "r", 0)
                        host_state = status_f.read().strip().upper()
                        status_f.close()
                        if host_state in ["ONLINE",
                                          "ACTIVE",
                                          "LINK",
                                          "LINKUP",
                                          ]:
                            break

                    except:
                        try:
                            status_f.close()
                        except:
                            pass

                    lipWait = lipWait - 1
                    time.sleep(1.0531)

                if lipWait <= 0:
                    IPlugin.logger.error("FC %s time out - Aborting!" % fc_host)
                    raise RefreshFailedEx(_("FC %s timed out!" % fc_host))

            else:
                IPlugin.logger.info("%s/issue_lip does not exist" % fc_path)

        except IOError, ioe:
            try:
                lip_f.close()
            except:
                pass

            IPlugin.logger.error("IO Error: %s" % ioe.strerror)

    for fc_host in os.listdir("/sys/class/scsi_host"):
        try:
            if os.path.exists("/sys/class/scsi_host/%s/scan" % fc_host):
                scan_f = open("/sys/class/scsi_host/%s/scan" % fc_host,
                          "w", 0)
                scan_f.write("- - -")
                scan_f.close()
            else:
                IPlugin.logger.info("/sys/class/scsi_host/%s/scan does not exist" % fc_host)
        except IOError, ioe:
            IPlugin.logger.error("IO Error: %s" % ioe.strerror)
            raise RefreshFailedEx(_("IO Error: %s" % ioe.strerror))


def mp_refresh():
    """
    Initiate a rescan for newly added or removed multipath LUNs.

    """
    from OSCPlugin import IPlugin

    if os.path.exists("/dev/emcpower"):
        pp_p = subprocess.Popen(["powermt", "config"],
                             stdout = subprocess.PIPE,
                             stderr = subprocess.PIPE,
                             close_fds=True)
        (p_out, p_err) = pp_p.communicate()
        if pp_p.returncode != 0:
            IPlugin.logger.error("Powerpath config failed: %s" % p_err)
    else:
        # Prevent multipath reload command on unmapped devices
        devices = OSCPluginUtils.getActiveDevMapper()
        for dev_name in devices:
            disk = os.path.join(IPlugin.dev_path_prefix, dev_name)
            mpt_p = subprocess.Popen(["multipath", "-r", disk],
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE,
                                 close_fds=True)
            (p_out, p_err) = mpt_p.communicate()
            if mpt_p.returncode != 0:
                IPlugin.logger.error("Multipath reload failed: %s" % p_err)

            devnull = open(os.devnull, "rw")

            subprocess.Popen(["dd", "oflag=nonblock", "if=%s" % disk, "of=/dev/null", "count=1"],
                              stderr=devnull, close_fds=True)
            devnull.close()


def mp_destroy(disk_id):
    """
    Tear down multipath device and associated slaves.

    """
    from OSCPlugin import IPlugin

    dev_name = "%s/%s" % (IPlugin.dev_path_prefix, disk_id)

    if not os.path.exists(dev_name):
        IPlugin.logger.warning("Device %s does not exist" % dev_name)
        return

    if os.path.exists("/dev/emcpower"):
        OSCPluginUtils.destroyPPDev(dev_name)

    else:
        OSCPluginUtils.destroyMPDev(dev_name)


def device_resize(ss_record, disk_id):
    """
    Refresh the size of a device.

    """
    from OSCPlugin import IPlugin
    from OSCPlugin import InvalidValueEx

    dev_name = "%s/%s" % (IPlugin.dev_path_prefix, disk_id)

    if not os.path.exists(dev_name):
        IPlugin.logger.error("Device %s does not exist" % dev_name)
        raise InvalidValueEx(_("Device %s does not exist" % dev_name))

    if os.path.exists("/dev/emcpower"):
        OSCPluginUtils.resizePPDev(ss_record, dev_name)
    else:
        OSCPluginUtils.resizeMPDev(dev_name)


def device_path(disk_id):
    """
    Return the device path of the device or None if the device
    does not exist.

    """
    from OSCPlugin import IPlugin

    dev_name = "%s/%s" % (IPlugin.dev_path_prefix, disk_id)
    if os.path.exists(dev_name):
        return dev_name
    else:
        return None


def iscsi_discovery(access_host, access_port, username, passwd, chap):

    from OSCPlugin import IPlugin
    from OSCPlugin import TargetDiscoveryEx

    if username or passwd or chap:
        OSCPluginUtils.iSCSIDiscoveryAuth(username, passwd, chap)
   
    host = socket.gethostbyname(access_host)
    portal = "%s:%s" % (host, access_port)

    target_list = []

    iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                               "--mode", "discovery",
                               "--type=sendtargets",
                               "--portal=%s" % portal],
                              stdout = subprocess.PIPE,
                              stderr = subprocess.PIPE,
                              close_fds=True)
    (p_out, p_err) = iscsi_p.communicate()
    if iscsi_p.returncode != 0:
        IPlugin.logger.error("Failure to get the targets: %s" % p_err)
        #raise TargetDiscoveryEx(_("Failure to get the targets: %s" % p_err))
        return target_list

    if os.path.exists("/etc/iscsi/iscsid.conf.bak"):
        os.remove("/etc/iscsi/iscsid.conf")
        os.rename("/etc/iscsi/iscsid.conf.bak",
                  "/etc/iscsi/iscsid.conf")

    # pylint: disable=E1103

    output =  p_out.splitlines()

    if p_out.startswith('Starting iscsid'):
        output = output[2:]

    for targets in output:
        target = targets.split(" ")
        dup = False
        if len(target) > 1:
            target_addr = target[0].split(',')[0]
            target_name = target[1] 

            if portal != target_addr:
                IPlugin.logger.info("Discovered target %s at address %s is not the same as the access portal %s and will be ignored" % 
                                        (target_name, target_addr, portal))
                continue

            for entry in target_list:
                if entry['iscsi_target_name'] == target_name and entry['iscsi_portal_address'] == host and entry['iscsi_portal_port'] == access_port:
                    dup = True
                    break
            if dup:
                continue
                        
            if username:
                iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                            "--mode", "node",
                                            "--targetname=%s" % target_name,
                                            "--op", "update",
                                            "--name=node.session.auth.username",
                                            "--value=%s" % username],
                                            stdout = subprocess.PIPE,
                                            stderr = subprocess.PIPE,
                                            close_fds=True)
                (p_out, p_err) = iscsi_p.communicate()
                if iscsi_p.returncode != 0:
                    IPlugin.logger.error("Unable to update username "
                                         "for %s: %s" % (target_name, p_err))
                    continue
                    #raise LoginFailedEx(_("Unable to update username "
                    #                      "for %s: %s" % (target_name, p_err)))

            if passwd:
                iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                            "--mode", "node",
                                            "--targetname=%s" % target_name,
                                            "--op", "update",
                                            "--name=node.session.auth.password",
                                            "--value=%s" % passwd],
                                            stdout = subprocess.PIPE,
                                            stderr = subprocess.PIPE,
                                            close_fds=True)
                (p_out, p_err) = iscsi_p.communicate()
                if iscsi_p.returncode != 0:
                    IPlugin.logger.error("Unable to update password "
                                         "for %s: %s" % (target_name, p_err))
                    continue
                    #raise LoginFailedEx(_("Unable to update password "
                    #                      "for %s: %s" % (target_name, p_err)))

            if chap:
                iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                            "--mode", "node",
                                            "--targetname=%s" % target_name,
                                            "--op", "update",
                                            "--name=node.session.auth.authmethod",
                                            "--value=CHAP"],
                                           stdout = subprocess.PIPE,
                                           stderr = subprocess.PIPE,
                                           close_fds=True)
                (p_out, p_err) = iscsi_p.communicate()
                if iscsi_p.returncode != 0:
                    IPlugin.logger.error("Unable to set CHAP "
                                         "authentication for "
                                         "%s: %s" % (target_name, p_err))
                    continue
                    #raise LoginFailedEx(_("Unable to set CHAP "
                    #                      "authentication for "
                    #                      "%s: %s" % (target_name, p_err)))


            target_entry = {}
            target_entry['iscsi_target_name'] = target_name
            target_entry['iscsi_portal_address'] = host
            target_entry['iscsi_portal_port'] = access_port
            target_list.append(target_entry)
       

        else:
            IPlugin.logger.error("Unable to determine target name")
            raise TargetDiscoveryEx(_("Unable to determine target name"))
            
    return target_list


def iscsi_login(username, passwd, chap, targets):
    if len(targets) > 50:
        portal = "%s:%s" % (targets[0]['iscsi_portal_address'], targets[0]['iscsi_portal_port']) 
        _login_target_all(portal, username, passwd, chap)
        return

    for target in targets:
        portal = "%s:%s" % (target['iscsi_portal_address'], target['iscsi_portal_port'])
        _login_target(portal, username, passwd, chap, target['iscsi_target_name'])



def _login_target_all(portal, username, passwd, chap):

    from OSCPlugin import IPlugin
    from OSCPlugin import LoginFailedEx



    if username:
        iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                    "--mode", "node",
                                  #  "--targetname=%s" % target,
                                    "--op", "update",
                                    "--name=node.session.auth.username",
                                    "--value=%s" % username],
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE,
                                   close_fds=True)
        (p_out, p_err) = iscsi_p.communicate()
        if iscsi_p.returncode != 0:
            IPlugin.logger.error("Unable to update username "
                                 "for %s: %s" % (portal, p_err))
            raise LoginFailedEx(_("Unable to update username "
                                  "for %s: %s" % (portal, p_err)))
    if passwd:
        iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                    "--mode", "node",
                                    #"--targetname=%s" % target,
                                    "--op", "update",
                                    "--name=node.session.auth.password",
                                    "--value=%s" % passwd],
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE,
                                   close_fds=True)
        (p_out, p_err) = iscsi_p.communicate()
        if iscsi_p.returncode != 0:
            IPlugin.logger.error("Unable to update password "
                                 "for %s: %s" % (portal, p_err))
            raise LoginFailedEx(_("Unable to update password "
                                  "for %s: %s" % (portal, p_err)))
    if chap:
        iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                    "--mode", "node",
                                    #"--targetname=%s" % target,
                                    "--op", "update",
                                    "--name=node.session.auth.authmethod",
                                    "--value=CHAP"],
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE,
                                   close_fds=True)
        (p_out, p_err) = iscsi_p.communicate()
        if iscsi_p.returncode != 0:
            IPlugin.logger.error("Unable to set CHAP "
                                 "authentication for "
                                 "%s: %s" % (portal, p_err))
            raise LoginFailedEx(_("Unable to set CHAP "
                                  "authentication for "
                                  "%s: %s" % (portal, p_err)))

    IPlugin.logger.debug("Loging into all targets portal %s" % portal)
    iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                "--mode", "node",
                               # "--targetname=%s" % target,
                                "--portal=%s" % portal,
                                "--login"],
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE,
                               close_fds=True)

    (p_out, p_err) = iscsi_p.communicate()
    if iscsi_p.returncode != 0:
        if iscsi_p.returncode != 15:
            raise LoginFailedEx(_("Target logins failed: %s" %
                                  p_err))


def _login_target(portal, username, passwd, chap, target):

    from OSCPlugin import IPlugin
    from OSCPlugin import LoginFailedEx

    if username:
        iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                    "--mode", "node",
                                    "--targetname=%s" % target,
                                    "--op", "update",
                                    "--name=node.session.auth.username",
                                    "--value=%s" % username],
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE,
                                   close_fds=True)
        (p_out, p_err) = iscsi_p.communicate()
        if iscsi_p.returncode != 0:
            IPlugin.logger.error("Unable to update username "
                                 "for %s: %s" % (target, p_err))
            raise LoginFailedEx(_("Unable to update username "
                                  "for %s: %s" % (target, p_err)))
    if passwd:
        iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                    "--mode", "node",
                                    "--targetname=%s" % target,
                                    "--op", "update",
                                    "--name=node.session.auth.password",
                                    "--value=%s" % passwd],
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE,
                                   close_fds=True)
        (p_out, p_err) = iscsi_p.communicate()
        if iscsi_p.returncode != 0:
            IPlugin.logger.error("Unable to update password "
                                 "for %s: %s" % (target, p_err))
            raise LoginFailedEx(_("Unable to update password "
                                  "for %s: %s" % (target, p_err)))
    if chap:
        iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                    "--mode", "node",
                                    "--targetname=%s" % target,
                                    "--op", "update",
                                    "--name=node.session.auth.authmethod",
                                    "--value=CHAP"],
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE,
                                   close_fds=True)
        (p_out, p_err) = iscsi_p.communicate()
        if iscsi_p.returncode != 0:
            IPlugin.logger.error("Unable to set CHAP "
                                 "authentication for "
                                 "%s: %s" % (target, p_err))
            raise LoginFailedEx(_("Unable to set CHAP "
                                  "authentication for "
                                  "%s: %s" % (target, p_err)))

    IPlugin.logger.debug("Logging in target %s, portal %s" % (target, portal))
    iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                "--mode", "node",
                                "--targetname=%s" % target,
                                "--portal=%s" % portal,
                                "--login"],
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE,
                               close_fds=True)

    (p_out, p_err) = iscsi_p.communicate()
    if iscsi_p.returncode != 0:
        if iscsi_p.returncode != 15:
            raise LoginFailedEx(_("Target login failed: %s" %
                                  p_err))


def iscsi_logout(targets):

    from OSCPlugin import IPlugin

    for target in targets:

        logout_cmd = [ IPlugin.iscsiadm,
                       "--mode", "node",
                       "--targetname=%s" % target['iscsi_target_name'],
                       "--portal=%s:%s" % (target['iscsi_portal_address'], target['iscsi_portal_port']),
                       "--logout"]

        iscsi_p = subprocess.Popen(logout_cmd,
                                   stdout = subprocess.PIPE,
                                   stderr = subprocess.PIPE,
                                   close_fds=True)

        (p_out, p_err) = iscsi_p.communicate()

        if iscsi_p.returncode != 0:
            IPlugin.logger.error("Target %s:%s:%s logout failed: %s" % 
                             (target['iscsi_target_name'], target['iscsi_portal_address'], str(target['iscsi_portal_port']), p_err))

        iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                "--mode", "node",
                                "--targetname=%s" % target['iscsi_target_name'],
                                "--portal=%s:%s" % (target['iscsi_portal_address'], target['iscsi_portal_port']),
                                "--op=delete"],
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE,
                               close_fds=True)

        (p_out, p_err) = iscsi_p.communicate()

        if iscsi_p.returncode != 0:
             IPlugin.logger.error("Failed to delete %s node record for portal: %s:%s" %
                             (target['iscsi_portal_address'], str(target['iscsi_portal_port']), p_err))



def iscsi_rescan(ss_record):

    from OSCPlugin import IPlugin
    from OSCPlugin import RefreshFailedEx

    
    iscsi_p = subprocess.Popen([IPlugin.iscsiadm,
                                "--mode", "node",
                                "--rescan"],
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE,
                               close_fds=True)

    (p_out, p_err) = iscsi_p.communicate()

    #Don't report no session errors.

    if iscsi_p.returncode != 0 and iscsi_p.returncode != 21:
        IPlugin.logger.error("Target refresh failed: %s" % p_err)
        raise RefreshFailedEx(_("Target refresh failed: %s" % p_err))

    time.sleep(1.2)
